#!/usr/local/bin/ruby -w
#
# vim:sw=2 ts=8:et sta:fdm=marker
#
#
# Copyright (c) 2005 Ariff Abdullah 
#        (skywizard@MyBSD.org.my) All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#        $MyBSD$
#
# Date: Sun Feb 20 20:49:23 MYT 2005
#   OS: FreeBSD kasumi.MyBSD.org.my 5.3-STABLE i386
#

require 'pathname'
require 'complex'

$:.unshift(File.dirname(Pathname.new(__FILE__).realpath()))

case RUBY_PLATFORM
  when /^i\d86\-(freebsd|linux)/i
    require $1.downcase() << '-ioccom'
else
  raise 'Unknown Platform'
end

def min(a, b)
  a < b ? a : b
end

class Numeric
  def clamp(min, max)
    if self < min
      min
    elsif self > max
      max
    else
      self
    end
  end
end

class WaveError < StandardError
end

class Wave
  HEADER_SIZE = 44
  HEADER_PACK = 'a4Ia4a4Is2I2s2a4I'
  def initialize(file, raw = nil, bits = nil, fmt = nil, rate = nil, channels = nil, skip = 0)
    @file = file.dup()
    @raw = raw
    if file.is_a?(IO) || file.is_a?(File)
      @fd = file
      @raw = true unless @raw == true || @raw == false
    else
      @fd = File.open(@file, 'rb')
    end
    if @raw
      @channels = channels
      @bits = bits
      @rate = rate
      @size = -1
      @skip = skip
      @fmt = fmt
      raise WaveError, 'Please supply sound format for raw playing' \
        unless @channels.is_a?(Integer) && @bits.is_a?(Integer) && \
          @rate.is_a?(Integer)
    else
      @fmt = @bits = @rate = @size = @channels = nil
      catch :success do
        begin
          while true
            hdr = @fd.sysread(8)
            raise WaveError, 'read failed' unless hdr.size == 8
            hdr, sz = hdr.unpack('a4I')
            case hdr
              when 'RIFF'
                hdr = @fd.sysread(4)
                raise WaveError, 'read failed' unless hdr.size == 4
                hdr = hdr.unpack('a4')[0]
                raise WaveError, 'Not a WAVE' unless hdr == 'WAVE'
              when 'fmt '
                raise WaveError, 'Illegal header' if sz < 16
                hdr = @fd.sysread(16)
                raise WaveError, 'read failed' unless hdr.size == 16
                waveheader = hdr.unpack('SSIISS')
                case waveheader[0]
                  when 0x0001
                    @fmt = waveheader[5]
                  when 0x0006, 0x0007
                    @fmt = waveheader[0]
                else
                  raise WaveError, 'Unsupported format 0x%04X' % waveheader[0]
                end
                @bits = waveheader[5]
                @rate = waveheader[2]
                @channels = waveheader[1]
                sz -= 16
                if sz > 0
                  pos = @fd.sysseek(0, IO::SEEK_CUR)
                  raise WaveError, 'seek failed' \
                      unless pos + sz == @fd.sysseek(sz, IO::SEEK_CUR)
                end
              when 'data'
                STDERR.puts 'Header size   : %d bytes' % @fd.sysseek(0, IO::SEEK_CUR)
                @size = sz
                throw :success
            else
              pos = @fd.sysseek(0, IO::SEEK_CUR)
              raise WaveError, 'Not a RIFF' if pos == 8
              raise WaveError, 'seek failed' \
                  unless pos + sz == @fd.sysseek(sz, IO::SEEK_CUR)
            end
          end
          raise EOFError
        rescue EOFError
          raise WaveError, 'data chunk not found'
        end
      end
      raise WaveError, 'Wave header error' \
          unless @bits && @rate && @channels && @fmt && @size
      @skip = 0
    end
    if @skip > 0
      left = @skip
      junk = ''
      readsz = 2352*16
      while left > 0
        junk.replace(@fd.sysread(min(readsz, left)))
        left -= junk.size()
      end
      junk = nil
      GC.start()
    end
    self
  end
  attr_reader :file, :channels, :bits, :rate, :size, :fmt
  def sysread(len)
    @fd.sysread(len)
  end
  def close
    @fd.close()
  end
end

class DSPError < StandardError
end

class DSP
  class << self
    include IOCCOM
  end
  AFMT_U8     = 0x00000008
  AFMT_S16_LE = 0x00000010
  AFMT_S16_BE = 0x00000020
  AFMT_S8     = 0x00000040
  AFMT_U16_LE = 0x00000080
  AFMT_U16_BE = 0x00000100
  AFMT_S32_LE = 0x00001000
  AFMT_S32_BE = 0x00002000
  AFMT_U32_LE = 0x00004000
  AFMT_U32_BE = 0x00008000
  AFMT_S24_LE = 0x00010000
  AFMT_S24_BE = 0x00020000
  AFMT_U24_LE = 0x00040000
  AFMT_U24_BE = 0x00080000
  AFMT_MU_LAW = 0x00000001
  AFMT_A_LAW  = 0x00000002
  FMT = {
    6 => [[AFMT_A_LAW, AFMT_A_LAW], [AFMT_A_LAW, AFMT_A_LAW]],
    7 => [[AFMT_MU_LAW, AFMT_MU_LAW], [AFMT_MU_LAW, AFMT_MU_LAW]],
    8 => [[AFMT_U8, AFMT_S8], [AFMT_U8, AFMT_S8]],
    16 => [[AFMT_S16_LE, AFMT_U16_LE], [AFMT_S16_BE, AFMT_U16_BE]],
    24 => [[AFMT_S24_LE, AFMT_U24_LE], [AFMT_S24_BE, AFMT_U24_BE]],
    32 => [[AFMT_S32_LE, AFMT_U32_LE], [AFMT_S32_BE, AFMT_U32_BE]]
  }
  FMT.default = [[AFMT_U8, AFMT_U8], [AFMT_U8, AFMT_U8]]
  sizeof_int = [0].pack('i').size()
  SNDCTL_DSP_CHANNELS      = _IOWR('P', 6, sizeof_int)
  SNDCTL_DSP_SPEED         = _IOWR('P', 2, sizeof_int)
  SNDCTL_DSP_SETFMT        = _IOWR('P', 5, sizeof_int)
  SNDCTL_DSP_RESET         = _IO('P', 0)
  SNDCTL_DSP_SYNC          = _IO('P', 1)
  SNDCTL_DSP_SETFRAGMENT   = _IOWR('P',10, sizeof_int)
  DEFAULT_DEV              = '/dev/dsp'
  BLOCKS                   = 512
  def initialize(dev = DEFAULT_DEV, si = true, le = true)
    @dev = dev.dup()
    @fd = File.open(@dev, 'wb')
    @si = si ? 0 : 1
    @le = le ? 0 : 1
  end
  def play(wv)
    STDOUT.sync = true
    STDERR.sync = true
    fmt = FMT[wv.fmt][@le][@si]
    raise DSPError, 'Failed to reset dsp' \
      if @fd.ioctl(SNDCTL_DSP_RESET) < 0
    raise DSPError, 'Failed to set dsp bit' \
      if @fd.ioctl(SNDCTL_DSP_SETFMT, [fmt].pack('i')) < 0
    raise DSPError, 'Failed to set dsp channel' \
      if @fd.ioctl(SNDCTL_DSP_CHANNELS, [wv.channels].pack('i')) < 0
    raise DSPError, 'Failed to set dsp rate' \
      if @fd.ioctl(SNDCTL_DSP_SPEED, [wv.rate].pack('i')) < 0
    bps = (wv.bits / 8) * wv.channels
    blksz = (bps * BLOCKS *  (wv.rate / 44100.0)).to_i
    blksz = bps if blksz < bps
    blksz -= blksz % bps
    frag = 0
    while (blksz >> frag) > 0
      frag += 1
    end
    frag |= 0xffff0000
    raise DSPError, 'Failed to set fragment' \
      if @fd.ioctl(SNDCTL_DSP_SETFRAGMENT, [frag].pack('i')) < 0
    buf = ''
    STDOUT.printf("Format        : 0x%08x\n", wv.fmt)
    STDOUT.puts "File name     : #{wv.file}"
    STDOUT.puts "Sampling rate : #{wv.rate} Hz"
    STDOUT.puts "Bits/Sample   : #{wv.bits} Bits"
    STDOUT.puts "Channels      : #{wv.channels}"
    STDOUT.puts "Block size    : #{bps} / #{blksz}"
    len = wv.size.to_s.size()
    total = wv.size
    ind = %w[- | / - \\]
    if total < 0
      total = nil
      throbber = 0
      fmt = "Size          : %s (unknown) Bytes\r"
      STDOUT.printf(fmt, ind[throbber % ind.size()])
      throbber += 1
    else
      throbber = -1
      fmt = "Size          : %#{len}d / #{wv.size} Bytes\r"
      STDOUT.printf(fmt, 0)
    end

    begin
      while true
        if throbber < 0
          buf << wv.sysread(min(blksz, total))
        else
          buf << wv.sysread(blksz)
        end
        # XXX face it, other than 16 bit, it sound like crapload
        until buf.empty?
          outlen = @fd.syswrite(buf.slice(0, blksz))
          buf.slice!(0, outlen)
          if throbber < 0
            total -= outlen
            STDOUT.printf(fmt, wv.size - total)
          else
            STDOUT.printf(fmt, ind[throbber % ind.size()])
            throbber += 1
          end
        end
        raise EOFError if total.is_a?(Integer) && total <= 0
      end
    rescue EOFError, Interrupt
    ensure
      @fd.ioctl(SNDCTL_DSP_SYNC)
    end
    STDOUT.print "\n"
  end
  def close
    @fd.ioctl(SNDCTL_DSP_SYNC)
    @fd.close()
  end
end

def usage(msg = nil)
  STDERR.puts msg if msg.is_a?(String)
  STDERR.puts "Usage: #{File.basename($0)} [-r -xs -xe] -C channels -B bits -R rate -s skipbytes files..."
  exit(1)
end

cfg = {
  :raw => false, :channels => 2, :bits => 16, :rate => 44100,
  :skip => 0, :sign => true, :le => true, :fmt => DSP::AFMT_S16_LE
}

token = nil
ARGV.delete_if do |arg|
  case arg
    when '-r'
      cfg[:raw] = true
      true
    when '-C'
      token = :channels
      true
    when '-B'
      token = :bits
      true
    when '-R'
      token = :rate
      true
    when '-s'
      token = :skip
      true
    when '-xs'
      cfg[:sign] = !cfg[:sign]
      true
    when '-xe'
      cfg[:le] = !cfg[:le]
      true
  else
    if token
      usage('Argument should be an Integer') unless arg.is_a?(String) \
          && arg =~ /^\d+$/
      arg = arg.to_i()
      if token == :bits
        if arg == DSP::AFMT_MU_LAW
          cfg[:fmt] = arg
          arg = 8;
        elsif arg == DSP::AFMT_A_LAW
          cfg[:fmt] = arg
          arg = 8
        else
          cfg[:fmt] = arg
        end
      end
      cfg[token] = arg
      token = nil
      true
    else
      false
    end
  end
end
usage('Option need an argument') if token
usage() if ARGV.empty?

dsp = DSP.new(DSP::DEFAULT_DEV, cfg[:sign], cfg[:le])
begin
  ARGV.each do |f|
    wv = nil
    begin
      if f == '-'
        wv = Wave.new(STDIN, cfg[:raw], cfg[:bits], cfg[:fmt], cfg[:rate], cfg[:channels], cfg[:skip])
      else
        if cfg[:raw]
          wv = Wave.new(f, cfg[:raw], cfg[:bits], cfg[:fmt], cfg[:rate], cfg[:channels], cfg[:skip])
        else
          wv = Wave.new(f)
        end
      end
      begin
        dsp.play(wv)
      rescue WaveError => err
        STDOUT.puts err.message
      end
    rescue Interrupt
      sleep(0.5)
    rescue WaveError => err
      STDERR.print f, ': ', err.message, "\n"
    rescue SystemCallError => err
      STDERR.puts err.message
    ensure
      wv.close() if wv && wv.file != STDIN
    end
  end
rescue Interrupt, SystemCallError
end
STDOUT.puts "Finish.."
dsp.close()
